import 'package:flutter/material.dart';
import 'package:apply/core/framework.dart';
import 'package:apply/core/plugin.dart';
import 'package:apply/elements/div.dart';
import 'package:apply/plugins/align_items.dart';
import 'package:apply/plugins/background_color.dart';
import 'package:apply/plugins/background_image.dart';
import 'package:apply/plugins/border_radius.dart';
import 'package:apply/plugins/border_width.dart';
import 'package:apply/plugins/box_shadow.dart';
import 'package:apply/plugins/display.dart';
import 'package:apply/plugins/flex_direction.dart';
import 'package:apply/plugins/flex_wrap.dart';
import 'package:apply/plugins/gradient_color_stops.dart';
import 'package:apply/plugins/grid.dart';
import 'package:apply/plugins/height.dart';
import 'package:apply/plugins/justify_content.dart';
import 'package:apply/plugins/margin.dart';
import 'package:apply/plugins/object-position.dart';
import 'package:apply/plugins/opacity.dart';
import 'package:apply/plugins/overflow.dart';
import 'package:apply/plugins/padding.dart';
import 'package:apply/plugins/position/position.dart';
import 'package:apply/plugins/space.dart';
import 'package:apply/plugins/width.dart';

class DisplayHandler {
  final Div element;

  Set<Plugin> get plugins => element.plugins;

  List<Widget> get children => element.children;

  final BuildContext context;

  DisplayHandler(this.element, this.context);

  WidgetBuilder run() {
    var baseWB = (_) => handleDisplay();
    WidgetBuilder result = baseWB;

    var buildPosition = _buildPosition();
    if (buildPosition != null) {
      var next = result;
      result = null;
      result = (_) => buildPosition(next);
    }

    var buildOverflow = _buildOverflow();
    if (buildOverflow != null) {
      var next = result;
      result = null;
      result = (_) => buildOverflow(next);
    }

    var buildOpacity = _buildOpacity();
    if (buildOpacity != null) {
      var next = result;
      result = null;
      result = (_) => buildOpacity(next);
    }

    return result;
  }

  Widget handleDisplay() {
    Widget result;
    Display display = plugins.getPluginByIdOrDefault("display");

    if (display.value == "flex") {
      FlexWrap flexWrap = plugins.getPluginById("flexWrap");

      if (flexWrap != null && !flexWrap.isNoWrap) {
        result = flexWrap.handle(element, context);
      } else {
        FlexDirection flexDirection = plugins.getPluginByIdOrDefault("flexDirection");
        result = flexDirection.handle(element, context);
      }
    } else if (display.value == "grid") {
      GridP gridPlugin = plugins.getPluginByIdOrDefault("grid");
      result = gridPlugin.handle(element, context);
    } else if (display.value == "hidden") {
      result = Offstage(
        offstage: true,
        child: Container(),
      );
      return result;
    } else if (display.value == "block") {
      if (children.length > 1) {
        result = Column(
          mainAxisSize: MainAxisSize.min,
        );
      } else if (children.length == 1) {
        result = this.children.first;
      }
    }

    if (result == null) {
      // print("Nema childova");
    }

    if (result is Flex) {
      Space space = plugins.getPluginById("space");
      if(space != null){
        space.handleFlex(element, context, (result as Flex).direction);
      }
      var mainAxisAlignment = _justifyContent();
      var crossAxisAlignment = _alignItems();
      result = (result as Flex).copyWith(
        children: this.children,
        mainAxisAlignment: mainAxisAlignment,
        crossAxisAlignment: crossAxisAlignment,
      );
    }

    ObjectPosition objectPosition = plugins.getPluginById("objectPosition");

    if (objectPosition != null) {
      result = objectPosition.getAlignedWidget(result);
    }

    return generateRootView(result);
  }

  Widget generateRootView(Widget child) {
    var options = element.divOptions;
    Color color = options.color;
    Decoration decoration = options.decoration;
    double width = options.width;
    double height = options.height;
    DecorationImage decorationImage;
    Gradient decorationGradient;
    BoxShadow boxShadow;
    BorderRadius borderRadius;
    Border border;

    EdgeInsets padding = options.padding;

    BackgroundColor backgroundColorPlugin = plugins.getPluginById("backgroundColor");

    if (backgroundColorPlugin != null) {
      color = backgroundColorPlugin.getColor(element);
    }

    PaddingP paddingPlugin = plugins.getPluginById("padding");

    BackgroundImage backgroundImage = plugins.getPluginById("backgroundImage");
    BoxShadowP boxShadowPlugin = plugins.getPluginById("boxShadow");
    BorderRadiusP borderRadiusPlugin = plugins.getPluginById("borderRadius");
    BorderWidth borderWidth = plugins.getPluginById("borderWidth");

    if (borderRadiusPlugin != null) {
      borderRadius = borderRadiusPlugin.getBorderRadius();
    }

    Width widthPlugin = plugins.getPluginById("width");
    Height heightPlugin = plugins.getPluginById("height");

    if (backgroundImage != null) {
      if (backgroundImage.modeOfOperation == BackgroundImageMode.gradient) {
        GradientColorStops gradientColorStops = plugins.getPluginById("gradientColorStops");
        if (gradientColorStops != null) {
          decorationGradient = gradientColorStops.getGradient(backgroundImage.getBeginEndPoints());
        }
      } else if (backgroundImage.modeOfOperation == BackgroundImageMode.assetImage) {
        decorationImage = backgroundImage.getAssetImageDecoration(element);
      }
    }

    if (boxShadowPlugin != null) {
      boxShadow = boxShadowPlugin.getShadow();
    }

    if (widthPlugin != null) {
      width = widthPlugin.getWidth();
    }

    if (heightPlugin != null) {
      height = heightPlugin.getHeight();
    }

    if (paddingPlugin != null) {
      padding = paddingPlugin.getPadding();
    }

    if (borderWidth != null) {
      border = borderWidth.getBorder(element);
    }

    if (decorationImage != null ||
        decorationGradient != null ||
        boxShadow != null ||
        borderRadius != null ||
        border != null) {
      decoration = BoxDecoration(
          boxShadow: boxShadow != null ? [boxShadow] : null,
          gradient: decorationGradient,
          image: decorationImage,
          color: color,
          borderRadius: borderRadius,
          border: border);
    }

    if (decoration != null) {
      color = null;
    }

    Widget container = Container(
      child: child,
      color: color,
      decoration: decoration,
      height: height,
      width: width,
      padding: padding,
    );

    if (width != null && width > 0 && width < 1) {
      double widthFactor = width;
      double heightFactor;
      if (height != null && height > 0 && height < 1) {
        heightFactor = height;
      }

      container = FractionallySizedBox(
        widthFactor: widthFactor,
        heightFactor: heightFactor,
        child: container,
      );
    }
    return container;
  }

  Widget addMargin(Widget child) {
    var options = element.divOptions;
    EdgeInsets margin = options.margin;
    Margin marginPlugin = plugins.getPluginById("margin");
    if (marginPlugin != null) {
      margin = marginPlugin.getMargin();
    } else {
      return child;
    }

    return margin != EdgeInsets.zero
        ? Padding(
            padding: margin,
            child: child,
          )
        : child;
  }

  ComposableWidget _buildOverflow() {
    OverflowP overflow = plugins.getPluginById("overflow");
    if (overflow == null) {
      return null;
    }
    var axis = overflow.getAxis();
    return (wb) => Builder(
          builder: (context) => SingleChildScrollView(
            scrollDirection: axis,
            child: wb(context),
          ),
        );
  }

  ComposableWidget _buildPosition() {
    Position position = plugins.getPluginById("position");
    if (position == null) {
      return null;
    }
    return (wb) => Builder(
          builder: (context) => position.handle(context, element, wb),
        );
  }

  ComposableWidget _buildOpacity() {
    OpacityP position = plugins.getPluginById("opacity");
    if (position == null) {
      return null;
    }
    return (wb) => Builder(
          builder: (context) => position.handle(context, element, wb),
        );
  }

  MainAxisAlignment _justifyContent() =>
      (plugins.getPluginByIdOrDefault("justifyContent") as JustifyContent).getAlignment();

  CrossAxisAlignment _alignItems() => (plugins.getPluginByIdOrDefault("alignItems") as AlignItems).getAlignment();

  WidgetBuilder call() => run();
}
